/********************************************************************************
 *   This file is part of JRtfTree.
 *
 *   JRtfTree is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   JRtfTree is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with JRtfTree; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ********************************************************************************/	

/********************************************************************************
 * Library:		JRtfTree
 * Version:     v0.1.1b
 * Class:		RtfLex
 * Copyright:   2005 Salvador Gomez
 * Home Page:	http://www.sgoliver.net
 * SF Project:	http://nrtftree.sourceforge.net
 *				http://sourceforge.net/projects/nrtftree
 * Date:		15/01/2006
 * Description:	Representa un documento RTF en forma de rbol.
 * ******************************************************************************/

package net.sgoliver.jrtftree.core;

import java.io.IOException;
import java.io.*;

/**
 * Analizador lxico (tokenizador) para documentos en formato RTF. Analiza el documento y devuelve de 
 * forma secuencial todos los elementos RTF leidos (tokens).
 */
public class RtfLex
{
	//Atributos
	private PushbackReader rtf;

	//Constantes
	private static final int EOF = -1;
	
	//Constructores
	
	/**
	 * Constructor de la clase RtfLex
	 * @param rtfreader Stream del fichero a analizar.
	 */
    public RtfLex(Reader rtfreader)
    {
        rtf = new PushbackReader(rtfreader);
    }
    
    //Mtodos Pblicos
    
    /**
     * Lee un nuevo token del documento RTF.
     */
    public RtfToken nextToken() throws IOException
    {
        //Caracter leido del documento
        int c;

        //Se crea el nuevo token a devolver
        RtfToken token = new RtfToken();

        //Se lee el siguiente caracter del documento
        c = rtf.read();

        //Se ignoran los retornos de carro, tabuladores y caracteres nulos
        while (c == '\r' || c == '\n' || c == '\t' || c == '\0')
            c = rtf.read();

        //Se trata el caracter leido
        if (c != EOF)
        {
            switch (c)
            {
                case '{':
                    token.tokenType = RtfTokenType.GROUP_START;
                    break;
                case '}':
                    token.tokenType = RtfTokenType.GROUP_END;
                    break;
                case '\\':
                    parseKeyword(token);
                    break;
                default:
                    token.tokenType = RtfTokenType.TEXT;
                    parseText(c, token);
                    break;
            }
        }
        else
        {
            //Fin de fichero
            token.tokenType = RtfTokenType.EOF;
        }

        return token;
    }
    
    //Mtodos Privados
    
    /**
     * Lee una cadena de Texto del documento RTF.
     * @param car Primer caracter de la cadena.
     * @param token Token RTF al que se asignar la palabra clave.
     */
    private void parseText(int car, RtfToken token) throws IOException
    {
        int c = car;

        StringBuffer texto = new StringBuffer(new String(new char[]{(char)c}));

        c = rtf.read();
        rtf.unread(c);

        //Se ignoran los retornos de carro, tabuladores y caracteres nulos
        while (c == '\r' || c == '\n' || c == '\t' || c == '\0')
        {
            rtf.read();
            
            //Peek reader
            c = rtf.read();
            rtf.unread(c);
        }

        while (c != '\\' && c != '}' && c != '{' && c != EOF)
        {
            rtf.read();

            texto.append((char)c);

            //Peek reader
            c = rtf.read();
            rtf.unread(c);

            //Se ignoran los retornos de carro, tabuladores y caracteres nulos
            while (c == '\r' || c == '\n' || c == '\t' || c == '\0')
            {
                rtf.read();
                
                //Peek reader
                c = rtf.read();
                rtf.unread(c);
            }
        }

        token.key = texto.toString();
    }
    
    /**
     * Lee una palabra clave del documento RTF.
     * @param token Token RTF al que se asignar la palabra clave.
     * @throws IOException
     */
    private void parseKeyword(RtfToken token) throws IOException
    {
        StringBuffer palabraClave = new StringBuffer();

        StringBuffer parametroStr = new StringBuffer();
        int parametroInt = 0;

        int c;
        boolean negativo = false;

        //Peek reader
        c = rtf.read();
        rtf.unread(c);

        //Si el caracter leido no es una letra --> Se trata de un smbolo de control
        if (!Character.isLetter((char)c))
        {
            rtf.read();
            token.tokenType = RtfTokenType.CONTROL;
            token.key = new String(new char[]{(char)c}); 

            //Si se trata de un caracter especial (codigo de 8 bits) se lee el parmetro hexadecimal
            if (token.key.equals("\'"))
            {
                String cod = "";

                cod += (char)rtf.read();
                cod += (char)rtf.read();

                token.hasParam = true;

                token.param = Integer.parseInt(cod,16); //Convert.ToInt32(cod, 16);
            }

            //Completar el resto de smbolos de control con parmetros.

            return;
        }

        //Se lee la palabra clave completa (hasta encontrar un caracter no alfanumrico, por ejemplo '\'  ' '
        
        //Peek reader
        c = rtf.read();
        rtf.unread(c);
        
        while (Character.isLetter((char)c))
        {
            rtf.read();
            palabraClave.append((char)c);

            //Peek reader
            c = rtf.read();
            rtf.unread(c);
        }

        //Se asigna la palabra clave leida
        token.tokenType = RtfTokenType.KEYWORD;
        token.key = palabraClave.toString();

        //Se comprueba si la palabra clave tiene parmetro
        if (Character.isDigit((char)c) || c == '-')
        {
            token.hasParam = true;

            //Se comprubea si el parmetro es negativo
            if (c == '-')
            {
                negativo = true;

                rtf.read();
            }

            //Se lee el parmetro completo
            
            //Peek reader
            c = rtf.read();
            rtf.unread(c);
            
            while (Character.isDigit((char)c))
            {
                rtf.read();
                parametroStr.append((char)c);

                //Peek reader
                c = rtf.read();
                rtf.unread(c);
            }

            parametroInt = Integer.parseInt(parametroStr.toString()); //Convert.ToInt32(parametroStr.ToString());

            if (negativo)
                parametroInt = -parametroInt;

            //Se asigna el parmetro de la palabra clave
            token.param = parametroInt;
        }

        if (c == ' ')
        {
            rtf.read();
        }
    }
}
