/********************************************************************************
 *   This file is part of JRtfTree.
 *
 *   JRtfTree is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   JRtfTree is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with JRtfTree; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ********************************************************************************/	

/********************************************************************************
 * Library:		JRtfTree
 * Version:     v0.1.1b
 * Class:		RtfTree
 * Copyright:   2005 Salvador Gomez
 * Home Page:	http://www.sgoliver.net
 * SF Project:	http://nrtftree.sourceforge.net
 *				http://sourceforge.net/projects/nrtftree
 * Date:		15/01/2006
 * Description:	Representa un documento RTF en forma de rbol.
 * ******************************************************************************/

package net.sgoliver.jrtftree.core;

import java.awt.Color;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.io.Reader;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.GregorianCalendar;

import net.sgoliver.jrtftree.util.InfoGroup;

/**
 * Reresenta la estructura en forma de rbol de un documento RTF. 
 */
public class RtfTree
{
	//Atributos
	private RtfTreeNode rootNode;
	private Reader rtf;
	private RtfLex lex;
	private RtfToken tok;
	private int level;
	
	/**
	 * Constructor de la clase. 
	 */
    public RtfTree()
    {
        //Se crea el nodo raz del documento
        rootNode = new RtfTreeNode(RtfNodeType.ROOT);

        //Se inicializa la profundidad actual
        level = 0;
    }
    
    /**
     * Carga un fichero en formato RTF.
     * @param path Ruta del fichero con el documento.
     * @return Se devuelve el valor 0 en caso de no producirse ningn error en la carga del documento. En caso contrario se devuelve el valor -1.
     */
    public int loadRtfFile(String path)
    {
        //Resultado de la carga
        int res = 0;

        try
        {
	        //Se abre el fichero de entrada
	        rtf = new FileReader(path);
	
	        //Se crea el analizador lxico para RTF
	        lex = new RtfLex(rtf);
	
	        //Se carga el rbol con el contenido del documento RTF
	        res = parseRtfTree();
	
	        //Se cierra el stream
	        rtf.close();
        }
        catch(IOException ex)
        {
        	res = -1;
        }

        //Se devuelve el resultado de la carga
        return res;
    }

    /**
     * Carga una cadena de Texto con formato RTF.
     * @param text Cadena de Texto que contiene el documento.
     * @return Se devuelve el valor 0 en caso de no producirse ningn error en la carga del documento. En caso contrario se devuelve el valor -1.
     */
    public int loadRtfText(String text)
    {
        //Resultado de la carga
        int res = 0;

        try
        {
	        //Se abre el fichero de entrada
	        rtf = new StringReader(text);
	
	        //Se crea el analizador lxico para RTF
	        lex = new RtfLex(rtf);
	
	        //Se carga el rbol con el contenido del documento RTF
	        res = parseRtfTree();
	
	        //Se cierra el stream
	        rtf.close();
	    }
	    catch(IOException ex)
	    {
	    	res = -1;
	    }

        //Se devuelve el resultado de la carga
        return res;
    }

    /**
     * Escribe el cdigo RTF del documento a un fichero.
     * @param filePath Ruta del fichero a generar con el documento RTF.
     * @throws IOException
     */
    public void saveRtf(String filePath) throws IOException
    { 
        //Stream de salida
        FileWriter sw = new FileWriter(filePath);

        //Se trasforma el rbol RTF a texto y se escribe al fichero
        sw.write(this.rootNode.getRtf());

        //Se cierra el fichero
        sw.flush();
        sw.close();
    }

    /**
     * Devuelve una representacin textual del documento cargado.
     */
    public String toString()
    {
        String res = "";

        res = toStringInm(rootNode, 0, false);

        return res;
    }

    /**
     * Devuelve una representacin textual del documento cargado. Aade el tipo de nodo a la izquierda del contenido del nodo.
     * @return Cadena de caracteres con la representacin del documento.
     */
    public String toStringEx()
    {
        String res = "";

        res = toStringInm(rootNode, 0, true);

        return res;
    }

    /**
     * Devuelve la tabla de fuentes del documento RTF.
     * @return Tabla de fuentes del documento RTF.
     */
    public String[] getFontTable()
    {
        ArrayList tabla = new ArrayList();
        String[] tablaFuentes;

		//Nodo raiz del documento
		RtfTreeNode root = this.rootNode;

		//Grupo principal del documento
		RtfTreeNode nprin = root.firstChild();

        //Buscamos la tabla de fuentes en el rbol
        boolean enc = false;
        int i = 0;
        RtfTreeNode ntf = new RtfTreeNode();  //Nodo con la tabla de fuentes

        while (!enc && i < nprin.getChildNodes().size())
        {
            if (nprin.getChildNodes().get(i).getNodeType() == RtfNodeType.GROUP &&
                nprin.getChildNodes().get(i).firstChild().getNodeKey().equals("fonttbl"))
            {
                enc = true;
                ntf = nprin.getChildNodes().get(i);
            }

            i++;
        }

        //Rellenamos el array de fuentes
        for (int j = 1; j < ntf.getChildNodes().size(); j++)
        {
            RtfTreeNode fuente = ntf.getChildNodes().get(j);

            String nombreFuente = null;

            for(int ind=0; ind<fuente.getChildNodes().size(); ind++)
            {
            	RtfTreeNode nodo = fuente.getChildNodes().get(i);
            	
                if (nodo.getNodeType() == RtfNodeType.TEXT)
                    nombreFuente = nodo.getNodeKey().substring(0, nodo.getNodeKey().length() - 1);
            }

            tabla.add(nombreFuente);
        }

        //Convertimos el ArrayList en un array tradicional
        tablaFuentes = new String[tabla.size()];

        for (int c = 0; c < tabla.size(); c++)
        {
            tablaFuentes[c] = (String)tabla.get(c);
        }

        return tablaFuentes;
    }

    /**
     * Devuelve la tabla de colores del documento RTF.
     * @return Tabla de colores del documento RTF.
     */
    public Color[] getColorTable()
    {
        ArrayList tabla = new ArrayList();
        Color[] tablaColores;

        //Nodo raiz del documento
        RtfTreeNode root = this.rootNode;

        //Grupo principal del documento
        RtfTreeNode nprin = root.firstChild();

        //Buscamos la tabla de colores en el rbol
        boolean enc = false;
        int i = 0;
        RtfTreeNode ntc = new RtfTreeNode();  //Nodo con la tabla de fuentes

        while (!enc && i < nprin.getChildNodes().size())
        {
            if (nprin.getChildNodes().get(i).getNodeType() == RtfNodeType.GROUP &&
                nprin.getChildNodes().get(i).firstChild().getNodeKey().equals("colortbl"))
            {
                enc = true;
                ntc = nprin.getChildNodes().get(i);
            }

            i++;
        }

        //Rellenamos el array de colores
        int rojo = 0;
        int verde = 0;
        int azul = 0;

        //Aadimos el color por defecto, en este caso el negro.
        //tabla.Add(Color.FromArgb(rojo,verde,azul));

        for (int j = 1; j < ntc.getChildNodes().size(); j++)
        {
            RtfTreeNode nodo = ntc.getChildNodes().get(j);

            if (nodo.getNodeType() == RtfNodeType.TEXT && nodo.getNodeKey().trim().equals(";"))
            {
                tabla.add(new Color(rojo, verde, azul));

                rojo = 0;
                verde = 0;
                azul = 0;
            }
            else if (nodo.getNodeType() == RtfNodeType.KEYWORD)
            {
                if(nodo.getNodeKey().equals("red"))
                	rojo = nodo.getParameter();
                else if(nodo.getNodeKey().equals("green"))
                    verde = nodo.getParameter();
                else if(nodo.getNodeKey().equals("blue"))
                    azul = nodo.getParameter();
            }
        }

        //Convertimos el ArrayList en un array tradicional
        tablaColores = new Color[tabla.size()];

        for (int c = 0; c < tabla.size(); c++)
        {
            tablaColores[c] = (Color)tabla.get(c);
        }

        return tablaColores;
    }

    /**
     * Devuelve la informacin contenida en el grupo "\info" del documento RTF.
     * @return Objeto InfoGroup con la informacin del grupo "\info" del documento RTF.
     */
    public InfoGroup getInfoGroup()
    {
        InfoGroup info = null;

        RtfTreeNode infoNode = this.rootNode.selectSingleNode("info");

        //Si existe el nodo "\info" exraemos toda la informacin.
        if (infoNode != null)
        {
            RtfTreeNode auxnode = null;

            info = new InfoGroup();

            //Title
            if ((auxnode = this.rootNode.selectSingleNode("title")) != null)
                info.setTitle(auxnode.nextSibling().getNodeKey());

            //Subject
            if ((auxnode = this.rootNode.selectSingleNode("subject")) != null)
                info.setSubject(auxnode.nextSibling().getNodeKey());

            //Author
            if ((auxnode = this.rootNode.selectSingleNode("author")) != null)
                info.setAuthor(auxnode.nextSibling().getNodeKey());

            //Manager
            if ((auxnode = this.rootNode.selectSingleNode("manager")) != null)
                info.setManager(auxnode.nextSibling().getNodeKey());

            //Company
            if ((auxnode = this.rootNode.selectSingleNode("company")) != null)
                info.setCompany(auxnode.nextSibling().getNodeKey());

            //Operator
            if ((auxnode = this.rootNode.selectSingleNode("operator")) != null)
                info.setOperator(auxnode.nextSibling().getNodeKey());

            //Category
            if ((auxnode = this.rootNode.selectSingleNode("category")) != null)
                info.setCategory(auxnode.nextSibling().getNodeKey());

            //Keywords
            if ((auxnode = this.rootNode.selectSingleNode("keywords")) != null)
                info.setKeywords(auxnode.nextSibling().getNodeKey());

            //Comments
            if ((auxnode = this.rootNode.selectSingleNode("comment")) != null)
                info.setComment(auxnode.nextSibling().getNodeKey());

            //Document comments
            if ((auxnode = this.rootNode.selectSingleNode("doccomm")) != null)
                info.setDoccomm(auxnode.nextSibling().getNodeKey());

            //Hlinkbase (The base address that is used for the path of all relative hyperlinks inserted in the document)
            if ((auxnode = this.rootNode.selectSingleNode("hlinkbase")) != null)
                info.setHlinkbase(auxnode.nextSibling().getNodeKey());

            //Version
            if ((auxnode = this.rootNode.selectSingleNode("version")) != null)
                info.setVersion(auxnode.getParameter());

            //Internal Version
            if ((auxnode = this.rootNode.selectSingleNode("vern")) != null)
                info.setInternalVersion(auxnode.getParameter());

            //Editing Time
            if ((auxnode = this.rootNode.selectSingleNode("edmins")) != null)
                info.setEditingTime(auxnode.getParameter());

            //Number of Pages
            if ((auxnode = this.rootNode.selectSingleNode("nofpages")) != null)
                info.setNofpages(auxnode.getParameter());

            //Number of Chars
            if ((auxnode = this.rootNode.selectSingleNode("nofchars")) != null)
                info.setNofchars(auxnode.getParameter());

            //Number of Words
            if ((auxnode = this.rootNode.selectSingleNode("nofwords")) != null)
                info.setNofwords(auxnode.getParameter());

            //Id
            if ((auxnode = this.rootNode.selectSingleNode("id")) != null)
                info.setId(auxnode.getParameter());

            //Creation DateTime
            if ((auxnode = this.rootNode.selectSingleNode("creatim")) != null)
                info.setCreationTime(parseDateTime(auxnode.getParentNode()));

            //Revision DateTime
            if ((auxnode = this.rootNode.selectSingleNode("revtim")) != null)
                info.setRevisionTime(parseDateTime(auxnode.getParentNode()));
            
            //Last Print Time
            if ((auxnode = this.rootNode.selectSingleNode("printim")) != null)
                info.setLastPrintTime(parseDateTime(auxnode.getParentNode()));

            //Backup Time
            if ((auxnode = this.rootNode.selectSingleNode("buptim")) != null)
                info.setBackupTime(parseDateTime(auxnode.getParentNode()));
        }

        return info;
    }

    //Mtodos Privados
    
    /**
     * Analiza el documento y lo carga con estructura de rbol.
     * @return Se devuelve el valor 0 en caso de no producirse ningn error en la carga del documento. En caso contrario se devuelve el valor -1.
     */
    private int parseRtfTree()
    {
        //Resultado de la carga del documento
        int res = 0;

        //Nodo actual
        RtfTreeNode curNode = rootNode;

        //Nuevos nodos para construir el rbol RTF
        RtfTreeNode newNode = null;

        try
        {
	        //Se obtiene el primer token
	        tok = lex.nextToken();
	
	        while (tok.tokenType != RtfTokenType.EOF)
	        {
	            switch (tok.tokenType)
	            {
	                case RtfTokenType.GROUP_START:
	                    newNode = new RtfTreeNode(RtfNodeType.GROUP);
	                    curNode.appendChild(newNode);
	                    curNode = newNode;
	                    level++;
	                    break;
	                case RtfTokenType.GROUP_END:
	                    curNode = curNode.getParentNode();
	                    level--;
	                    break;
	                case RtfTokenType.KEYWORD:
	                case RtfTokenType.CONTROL:
	                case RtfTokenType.TEXT:
	                    newNode = new RtfTreeNode(tok);
	                    curNode.appendChild(newNode);
	                    break;
	                default:
	                    res = -1;
	                    break;
	            }
	
	            //Se obtiene el siguiente token
	            tok = lex.nextToken();
	        }
	        
	        //Si el nivel actual no es 0 ( == Algun grupo no est bien formado )
	        if (level != 0)
	        {
	            res = -1;
	        }
        }
        catch(IOException ex)
        {
        	res = -1;
        }

        //Se devuelve el resultado de la carga
        return res;
    }

    /**
     * Mtodo auxiliar para generar la representacin Textual del documento RTF.
     * @param curNode Nodo actual del rbol.
     * @param level Nivel actual en rbol.
     * @param showNodeTypes Indica si se mostrar el tipo de cada nodo del rbol.
     * @return Representacin Textual del nodo 'curNode' con nivel 'level'.
     */
    private String toStringInm(RtfTreeNode curNode, int level, boolean showNodeTypes)
    {
        StringBuffer res = new StringBuffer();

        RtfNodeCollection children = curNode.getChildNodes();

        for (int i = 0; i < level; i++)
            res.append("  ");

        if (curNode.getNodeType() == RtfNodeType.ROOT)
            res.append("ROOT\r\n");
        else if (curNode.getNodeType() == RtfNodeType.GROUP)
            res.append("GROUP\r\n");
        else
        {
            if (showNodeTypes)
            {
                res.append(RtfNodeType.toString(curNode.getNodeType()));
                res.append(": ");
            }

            res.append(curNode.getNodeKey());

            if (curNode.getHasParameter())
            {
                res.append(" ");
                res.append(String.valueOf(curNode.getParameter()));
            }

            res.append("\r\n");
        }

        for(int i=0; i<children.size(); i++)
        {
        	RtfTreeNode node = children.get(i); 
            res.append(toStringInm(node, level + 1, showNodeTypes));
        }

        return res.toString();
    }

    /**
     * Parsea una fecha con formato "\yr2005\mo12\dy2\hr22\min56\sec15"
     * @param group Grupo RTF con la fecha.
     * @return Objeto DateTime con la fecha leida.
     */
    private Calendar parseDateTime(RtfTreeNode group)
    {
        Calendar dt = null;

        int year = 0, month = 0, day = 0, hour = 0, min = 0, sec = 0;

        for(int i=0; i<group.getChildNodes().size(); i++)
        {
        	RtfTreeNode node =  group.getChildNodes().get(i);
        	
            if(node.getNodeKey().equals("yr"))
          		year = node.getParameter();
            else if(node.getNodeKey().equals("mo"))
                month = node.getParameter();
            else if(node.getNodeKey().equals("dy"))
                day = node.getParameter();
            else if(node.getNodeKey().equals("hr"))
                hour = node.getParameter();
            else if(node.getNodeKey().equals("min"))
                min = node.getParameter();
            else if(node.getNodeKey().equals("sec"))
                sec = node.getParameter();
        }

        dt = new GregorianCalendar(year, month, day, hour, min, sec);

        return dt;
    }
    
    //Propiedades
    
    /**
     * Obtiene el nodo raz del rbol RTF.
     */
    public RtfTreeNode getRootNode()
    {
    	return rootNode;
    }
    
    /**
     * Obtiene el cdigo RTF completo del documento.
     * @return Cdigo RTF del documento completo.
     */
    public String getRtf()
    {
    	return rootNode.getRtf();
    }
}
