/********************************************************************************
 *   This file is part of JRtfTree.
 *
 *   JRtfTree is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   JRtfTree is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with JRtfTree; if not, write to the Free Software
 *   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
 ********************************************************************************/	

/********************************************************************************
 * Library:		JRtfTree
 * Version:     v0.1.1b
 * Class:		RtfTreeNode
 * Copyright:   2005 Salvador Gomez
 * Home Page:	http://www.sgoliver.net
 * SF Project:	http://nrtftree.sourceforge.net
 *				http://sourceforge.net/projects/nrtftree
 * Date:		15/01/2006
 * Description:	Representa un documento RTF en forma de rbol.
 * ******************************************************************************/

package net.sgoliver.jrtftree.core;

/**
 * Nodo RTF de la representacin en rbol de un documento. 
 */
public class RtfTreeNode
{
	//Atributos
	protected int nodeType;
	protected String key;
	protected boolean hasParam;
	protected int param;
	protected RtfNodeCollection children;
	protected RtfTreeNode parent;
	protected RtfTreeNode root;
	
	//Constructores
	
	/**
	 * Constructor por defecto de la clase.
	 */
    public RtfTreeNode()
    {
        children = new RtfNodeCollection();

        this.nodeType = RtfNodeType.NONE;
        this.key = "";
        this.hasParam = false;
        this.param = 0;

        this.parent = null;
        this.root = null;
    }

    /**
     * Constructor de la clase. Crea un nodo vaco del tipo pasado como parmetro.
     * @param nodeType Tipo de nodo a crear.
     */
    public RtfTreeNode(int nodeType)
    {
        children = new RtfNodeCollection();

        this.nodeType = nodeType;
        this.key = "";
        this.hasParam = false;
        this.param = 0;

        this.parent = null;
        this.root = null;
    }

    /**
     * Constructor de la clase. Crea un nodo completo con los datos suministrados como parametros.
     * @param nodeType Tipo de nodo.
     * @param key Clave del nodo.
     * @param hasParam Indicativo de existencia de parmetro.
     * @param param Parmetro del nodo, en caso de existir.
     */
    public RtfTreeNode(int nodeType, String key, boolean hasParam, int param)
    {
        children = new RtfNodeCollection();

        this.nodeType = nodeType;
        this.key = key;
        this.hasParam = hasParam;
        this.param = param;

        this.parent = null;
        this.root = null;
    }
    
    /**
     * Constructor privado de la clase. Crea un nodo a partir de un token del analizador lxico.
     * @param token Token RTF devuelto por el analizador lxico.
     */
    protected RtfTreeNode(RtfToken token)
    {
        children = new RtfNodeCollection();

        this.nodeType = token.tokenType;  //Ojo!
        this.key = token.key;
        this.hasParam = token.hasParam;
        this.param = token.param;

        this.parent = null;
        this.root = null;
    }
    
    //Mtodos Pblicos
    
    /**
     * Aade un nodo al final de la lista de hijos.
     * @param newNode Nuevo nodo a aadir.
     */
    public void appendChild(RtfTreeNode newNode)
    {
        //Se asigna como nodo padre el nodo actual
        newNode.parent = this;

        //Se asigna el nodo raz del documento
        newNode.root = this.root;

        //Se aade el nuevo nodo al final de la lista de nodos hijo
        children.add(newNode);
    }

    /**
     * Elimina un nodo de la lista de hijos.
     * @param index Indice del nodo a eliminar.
     */
    public void removeChild(int index)
    {
        //Se elimina el i-simo hijo
        children.remove(index);
    }

    /**
     * Elimina un nodo de la lista de hijos.
     * @param node Nodo a eliminar.
     */
    public void removeChild(RtfTreeNode node)
    {
        //Se busca el nodo a eliminar
        int index = children.indexOf(node);

        //Se elimina el i-simo hijo
        children.remove(index);
    }

    /**
     * Realiza una copia exacta del nodo actual.
     * @param cloneChildren Si este parmetro recibe el valor true se clonarn tambin todos los nodos hijo del nodo actual.
     * @return Devuelve una copia exacta del nodo actual.
     */
    public RtfTreeNode cloneNode(boolean cloneChildren)
    {
        RtfTreeNode clon = new RtfTreeNode();

        clon.key = this.key;
        clon.hasParam = this.hasParam;
        clon.param = this.param;
        clon.parent = this.parent;
        clon.root = this.root;
        clon.nodeType = this.nodeType;

        //Si cloneChildren=false se copia directamente la lista de hijos
        if(!cloneChildren)
        {
            clon.children = this.children;
        }
        else  //En caso contrario se clonan tambin cada uno de los hijos, propagando el parmetro cloneChildren=true
        {
            clon.children = new RtfNodeCollection();

            for(int i=0; i<children.size(); i++)
            {
            	clon.children.add(children.get(i).cloneNode(true));
            }
        }

        return clon;
    }

    /**
     * Indica si el nodo actual tiene nodos hijos.
     * @return Devuelve true si el nodo actual tiene algn nodo hijo.
     */
    public boolean hasChildNodes()
    {
        return (children.size() != 0);
    }

    /**
     * Devuelve el primer nodo de la lista de nodos hijos del nodo actual cuya palabra clave es la indicada como parmetro.
     * @param keyword Palabra clave buscada.
     * @return Primer nodo de la lista de nodos hijos del nodo actual cuya palabra clave es la indicada como parmetro.
     */
    public RtfTreeNode selectSingleChildNode(String keyword)
    {
        int i = 0;
        boolean found = false;
        RtfTreeNode node = null;

        while (i < children.size() && !found)
        {
            if (children.get(i).key.equals(keyword))
            {
                node = children.get(i);
                found = true;
            }

            i++;
        }

        return node;
    }

    /**
     * Devuelve el primer nodo de la lista de nodos hijos del nodo actual cuyo tipo es el indicado como parmetro.
     * @param nodeType Tipo de nodo buscado.
     * @return Primer nodo de la lista de nodos hijos del nodo actual cuyo tipo es el indicado como parmetro.
     */
    public RtfTreeNode selectSingleChildNodeByType(int nodeType)
    {
        int i = 0;
        boolean found = false;
        RtfTreeNode node = null;

        while (i < children.size() && !found)
        {
            if (children.get(i).nodeType == nodeType)
            {
                node = children.get(i);
                found = true;
            }

            i++;
        }

        return node;
    }

    /**
     * Devuelve el primer nodo del rbol, a partir del nodo actual, cuyo tipo es el indicado como parmetro.
     * @param nodeType Tipo del nodo buscado.
     * @return Primer nodo del rbol, a partir del nodo actual, cuyo tipo es el indicado como parmetro.
     */
    public RtfTreeNode selectSingleNodeByType(int nodeType)
    {
        int i = 0;
        boolean found = false;
        RtfTreeNode node = null;

        while (i < children.size() && !found)
        {
            if (children.get(i).nodeType == nodeType)
            {
                node = children.get(i);
                found = true;
            }
            else
            {
                node = children.get(i).selectSingleNodeByType(nodeType);

                if (node != null)
                {
                    found = true;
                }
            }

            i++;
        }

        return node;
    }

    /**
     * Devuelve el primer nodo del rbol, a partir del nodo actual, cuya palabra clave es la indicada como parmetro.
     * @param keyword Palabra clave buscada.
     * @return Primer nodo del rbol, a partir del nodo actual, cuya palabra clave es la indicada como parmetro.
     */
    public RtfTreeNode selectSingleNode(String keyword)
    {
        int i = 0;
        boolean found = false;
        RtfTreeNode node = null;

        while (i < children.size() && !found)
        {
            if (children.get(i).key.equals(keyword))
            {
                node = children.get(i);
                found = true;
            }
            else
            {
                node = children.get(i).selectSingleNode(keyword);

                if (node != null)
                {
                    found = true;
                }
            }

            i++;
        }

        return node;
    }

    /**
     * Devuelve todos los nodos, a partir del nodo actual, cuya palabra clave es la indicada como parmetro.
     * @param keyword Palabra clave buscada.
     * @return Coleccin de nodos, a partir del nodo actual, cuya palabra clave es la indicada como parmetro.
     */
    public RtfNodeCollection selectNodes(String keyword)
    {
        RtfNodeCollection nodes = new RtfNodeCollection();

        for(int i=0; i<children.size(); i++)
        {
            if (children.get(i).key.equals(keyword))
            {
                nodes.add(children.get(i));
            }

            nodes.addAll(children.get(i).selectNodes(keyword));
        }

        return nodes;
    }

    /**
     * Devuelve todos los nodos, a partir del nodo actual, cuyo tipo es el indicado como parmetro.
     * @param nodeType Tipo del nodo buscado.
     * @return Coleccin de nodos, a partir del nodo actual, cuyo tipo es la indicado como parmetro.
     */
    public RtfNodeCollection selectNodesByType(int nodeType)
    {
        RtfNodeCollection nodes = new RtfNodeCollection();

        for(int i=0; i<children.size(); i++)
        {
            if (children.get(i).nodeType == nodeType)
            {
                nodes.add(children.get(i));
            }

            nodes.addAll(children.get(i).selectNodesByType(nodeType));
        }

        return nodes;
    }

    /**
     * Devuelve todos los nodos de la lista de nodos hijos del nodo actual cuya palabra clave es la indicada como parmetro.
     * @param keyword Palabra clave buscada.
     * @return Coleccin de nodos de la lista de nodos hijos del nodo actual cuya palabra clave es la indicada como parmetro.
     */
    public RtfNodeCollection selectChildNodes(String keyword)
    {
        RtfNodeCollection nodes = new RtfNodeCollection();

        for(int i=0; i<children.size(); i++)
        {
            if (children.get(i).key.equals(keyword))
            {
                nodes.add(children.get(i));
            }
        }

        return nodes;
    }

    /**
     * Devuelve todos los nodos de la lista de nodos hijos del nodo actual cuyo tipo es el indicado como parmetro.
     * @param nodeType Tipo del nodo buscado.
     * @return Coleccin de nodos de la lista de nodos hijos del nodo actual cuyo tipo es el indicado como parmetro.
     */
    public RtfNodeCollection selectChildNodesByType(int nodeType)
    {
        RtfNodeCollection nodes = new RtfNodeCollection();

        for(int i=0; i<children.size(); i++)
        {
            if (children.get(i).nodeType == nodeType)
            {
                nodes.add(children.get(i));
            }
        }

        return nodes;
    }
    
    //Mtodos Privados
    
    /**
     * Mtodo auxiliar para obtener el Texto RTF del nodo actual a partir de su representacin en rbol.
     * @param curNode Nodo actual del rbol.
     * @param prevNode Nodo anterior tratado.
     * @return Texto en formato RTF del nodo. 
     */
    private String getRtfInm(RtfTreeNode curNode, RtfTreeNode prevNode)
    {
    	StringBuffer res = new StringBuffer("");

        if (curNode.nodeType == RtfNodeType.ROOT)
            res.append("");
        else if (curNode.nodeType == RtfNodeType.GROUP)
            res.append("{");
        else
        {
            if (curNode.nodeType != RtfNodeType.TEXT)
            {
                res.append("\\");
            }
            else  //curNode.NodeType == RTF_NODE_TYPE.TEXT
            {
                if (prevNode == null || prevNode.nodeType == RtfNodeType.CONTROL)
                {
                    res.append("");
                }
                else //antNode.NodeType == RTF_NODE_TYPE.KEYWORD
                {
                    res.append(" ");
                }
            }

            res.append(curNode.key);

            if (curNode.hasParam)
            {
                if (curNode.nodeType == RtfNodeType.KEYWORD)
                {
                    res.append(String.valueOf(curNode.param));
                }
                else if (curNode.nodeType == RtfNodeType.CONTROL)
                {
                    //Si es un caracter especial como las vocales acentuadas
                    if (curNode.key.equals("\'"))
                    {
                    	res.append(Integer.toHexString(curNode.param));
                    }
                }
            }
        }

        //Se obtienen los nodos hijos
        RtfNodeCollection children = curNode.children;

        for (int i = 0; i < children.size(); i++)
        {
            RtfTreeNode node = children.get(i);

            if (i > 0)
                res.append(getRtfInm(node, children.get(i-1)));
            else
                res.append(getRtfInm(node, null));

            if (node.nodeType == RtfNodeType.GROUP)
            {
                res.append("}");
            }
        }

        return res.toString();
    }
    
    /**
     * Obtiene el nodo raz del rbol RTF.
     * @return Nodo raz del rbol RTF.
     */
    public RtfTreeNode getRootNode()
    {
    	return root;
    }
    
    /**
     * Establece el nodo raz del rbol RTF.
     * @param node Nodo raz del rbol RTF.
     */
    public void setRootNode(RtfTreeNode node)
    {
    	root = node;
    }
    
    /**
     * Obtiene el nodo padre del nodo actual.
     * @return Nodo padre del nodo actual.
     */
    public RtfTreeNode getParentNode()
    {
    	return parent;
    }    
    
    /**
     * Establece el nodo padre del nodo actual.
     * @param node Nodo padre del nodo actual.
     */
    public void setParentNode(RtfTreeNode node)
    {
    	parent = node;
    }
    
    /**
     * Obtiene el tipo de nodo actual.
     * @return Tipo del nodo actual.
     */
    public int getNodeType()
    {
    	return nodeType;
    }
    
    /**
     * Establece el tipo de nodo actual.
     * @param nodeType Tipo de nodo.
     */
    public void setNodeType(int nodeType)
    {
    	this.nodeType = nodeType; 
    }
    
    /**
     * Obtiene la clave del nodo actual.
     * @return Clave del nodo actual.
     */
    public String getNodeKey()
    {
    	return key;
    }
    
    /**
     * Establece la clave del nodo actual.
     * @param key Clave del nodo.
     */
    public void setNodeKey(String key)
    {
    	this.key = key;
    }
    
    /**
     * Obtiene el indicativo de existencia de parmetro asociado al nodo.
     * @return Indicativo de existencia de parmetro asociado al nodo.
     */
    public boolean getHasParameter()
    {
    	return hasParam;
    }
    
    /**
     * Establece el indicativo de existencia de parmetro asociado al nodo.
     * @param hasParam Indicativo de existencia de parmetro asociado al nodo.
     */
    public void setHasParamenter(boolean hasParam)
    {
    	this.hasParam = hasParam;
    }    
    
    /**
     * Obtiene el parmetro asociado al nodo.
     * @return Parmetro asociado al nodo.
     */
    public int getParameter()
    {
    	return param;
    }
    
    /**
     * Establece el parmetro asociado al nodo.
     * @param param Parmetro asociado al nodo.
     */
    public void setParameter(int param)
    {
    	this.param = param; 
    }
    
    /**
     * Obtienen la coleccin de nodos hijo del nodo actual.
     * @return Coleccin de nodos hijo del nodo actual.
     */
    public RtfNodeCollection getChildNodes()
    {
    	return children;
    }
    
    /**
     * Establece la coleccin de nodos hijo del nodo actual.
     * @param children Coleccin de nodos hijo del nodo actual.
     */
    public void setChildNodes(RtfNodeCollection children)
    {
    	this.children = children;
    }
    
    /**
     * Obtiene el primer nodo hijo del nodo actual.
     * @return Primer nodo hijo del nodo actual.
     */
    public RtfTreeNode firstChild()
    {
        if (children.size() > 0)
            return children.get(0);
        else
            return null;
    }
    
    /**
     * Obtiene el ltimo nodo hijo del nodo actual.
     * @return ltimo nodo hijo del nodo actual.
     */
    public RtfTreeNode lastChild()
    {
    	if (children.size() > 0)
            return children.get(children.size()-1);
        else
            return null;
    }
    
    /**
     * Obtiene el nodo hermano siguiente del nodo actual (Dos nodos son hermanos si tienen el mismo nodo padre [ParentNode]).
     * @return Siguiente nodo hermano del nodo actual.
     */
    public RtfTreeNode nextSibling()
    {
        int currentIndex = parent.children.indexOf(this);

        if (parent.children.size() > currentIndex + 1)
            return parent.children.get(currentIndex + 1);
        else
            return null;
    }
    
    /**
     * Obtiene el nodo hermano anterior del nodo actual (Dos nodos son hermanos si tienen el mismo nodo padre [ParentNode]).
     * @return Nodo hermano anterior del nodo actual.
     */
    public RtfTreeNode previousSibling()
    {
        int currentIndex = parent.children.indexOf(this);

        if (currentIndex > 0)
            return parent.children.get(currentIndex - 1);
        else
            return null;
    }
    
    /**
     * Devuelve el cdigo RTF del nodo actual y todos sus nodos hijos.
     * @return Cdigo RTF del nodo actual y todos sus nodos hijos.
     */
    public String getRtf()
    {
        String res = "";

        res = getRtfInm(this, null);

        return res;
    }
}
